#include <stdlib.h>
#include <string.h>

#include "block.h"
#include "area.h"

typedef struct _AreaPiece *AreaPiece;

struct _AreaPiece {
  struct _AreaPiece *next;
  struct _AreaPiece *prev;
  int size;
  char data[0];
};

struct _Area {
  BlockPool pool;
  int size;
  struct _AreaPiece head;
  struct _AreaPiece tail;
};

Area Area_Create(BlockPool pool)
{
  Area area;

  area = BlockPool_Allocate(pool, sizeof(*area));

  area->pool = pool;
  area->size = 0;

  area->head.prev = NULL;
  area->head.next = &area->tail;
  area->tail.prev = &area->head;
  area->tail.next = NULL;

  area->head.size = 0;
  area->tail.size = 0;

  return area;
}

Area Area_Destroy(Area area)
{
  AreaPiece piece, next;

  if (area) {
    for (piece = area->head.next; piece != &area->tail; piece = next) {
      next = piece->next;
      BlockPool_Free(area->pool, piece);
    }
    BlockPool_Free(area->pool, area);
  }

  return NULL;
}

BlockPool Area_GetPool(Area area)
{
  return area->pool;
}

int Area_GetSize(Area area)
{
  return area->size;
}

enum {
  AREA_ACCESS_MODE_READ,
  AREA_ACCESS_MODE_WRITE,
  AREA_ACCESS_MODE_COMPARE,
};

static int Area_Access(Area area, int offset, int size,
		       const void *read_buffer, void *write_buffer, int mode)
{
  AreaPiece piece;
  int s, ret = 0, done = 0, cmp = 0;

  for (piece = area->head.next; piece != &area->tail; piece = piece->next) {
    if (size <= 0) break;
    s = piece->size;
    if (offset >= s) {
      offset -= s;
    } else {
      s -= offset;
      if (s > size) s = size;

      switch (mode) {
      case  AREA_ACCESS_MODE_READ:
	memcpy((char *)write_buffer + done, piece->data + offset, s);
	break;
      case  AREA_ACCESS_MODE_WRITE:
	memcpy(piece->data + offset, (char *)read_buffer + done, s);
	break;
      case  AREA_ACCESS_MODE_COMPARE:
	if (memcmp(piece->data + offset, (char *)read_buffer + done, s))
	  cmp = 1;
	break;
      default:
	break;
      }

      done += s;
      size -= s;
      offset = 0;
    }
  }

  switch (mode) {
  case  AREA_ACCESS_MODE_READ:    ret = done; break;
  case  AREA_ACCESS_MODE_WRITE:   ret = done; break;
  case  AREA_ACCESS_MODE_COMPARE: ret = (size > 0) ? -1 : cmp; break;
  default: break;
  }

  return ret;
}

int Area_Read(Area area, int offset, int size, void *buffer)
{
  return Area_Access(area, offset, size, NULL, buffer, AREA_ACCESS_MODE_READ);
}

int Area_Write(Area area, int offset, int size, const void *buffer)
{
  return Area_Access(area, offset, size, buffer, NULL, AREA_ACCESS_MODE_WRITE);
}

int Area_Compare(Area area, int offset, int size, const void *buffer)
{
  return Area_Access(area, offset, size, buffer, NULL, AREA_ACCESS_MODE_COMPARE);
}

int Area_Clear(Area area)
{
  return Area_Resize(area, 0);
}

/* ꤷĥ */
int Area_Extend(Area area, int size)
{
  AreaPiece piece;

  piece = BlockPool_Allocate(area->pool, sizeof(*piece) + size);

  piece->next = &area->tail;
  piece->prev = area->tail.prev;
  piece->size = size;

  area->tail.prev->next = piece;
  area->tail.prev = piece;
  area->size += size;

  return area->size;
}

/* ꤷ̾ */
int Area_Reduce(Area area, int size)
{
  AreaPiece piece;

  if (size > area->size) size = area->size;

  while (size > 0) {
    piece = area->tail.prev;
    if (size < piece->size) {
      area->size -= size;
      piece->size -= size;
      size = 0;
    } else {
      area->size -= piece->size;
      piece->prev->next = piece->next;
      piece->next->prev = piece->prev;
      size -= piece->size;
      BlockPool_Free(area->pool, piece);
    }
  }

  return area->size;
}

int Area_Resize(Area area, int size)
{
  if (size > area->size)
    return Area_Extend(area, size - area->size);
  else
    return Area_Reduce(area, area->size - size);
}
